
from logging import getLogger
import traceback
from os import getenv
from botocore.exceptions import ClientError
import sqlalchemy as sa
from sqlalchemy.engine.url import URL
from sqlalchemy import orm as sa_orm
from math import ceil
from nanoid import generate
import sqlalchemy as sa
from sqlalchemy.engine.url import URL
from klp_commons.log.structuredMessage import StructuredMessage
from sqlalchemy import orm as sa_orm
from datetime import datetime
from pandas import read_sql_query
from sqlalchemy import text
from datetime import datetime,timedelta
from pandas import DataFrame
from pandas import concat
from numpy import percentile
from numpy import arange
from json import loads as loads_
#import uuid
#import random

fmt = StructuredMessage
message = 'redshiftController'


class ControllerRedShift:
    def __init__(self):

        self.error = False
        self.url_string= None
        self.metadata = None
        self.session = None
        self.engine = None
        self.region_name = None
        self.driver = 'redshift+redshift_connector'
        self.REDSHIFT_DATABASE = getenv("REDSHIFT_DATABASE")
        self.REDSHIFT_USER = getenv("REDSHIFT_USER")
        self.REDSHIFT_PASSWORD = getenv("REDSHIFT_PASSWORD")
        self.REDSHIFT_HOST = getenv("REDSHIFT_HOST")
        self.REDSHIFT_PORT = getenv("REDSHIFT_PORT")
        self.prefix = getenv("NODE_ENV") + '_'
        self.date_format= '%Y-%m-%dT%H:%M:%S'

        # set self.url_string
        self.get_url()

        # module_name.TopicKafka_name
        self.logger = getLogger(message)

        # Init ()
        self.logger.info("donde init method of ControllerRedShift ")

        """
        engine = sa.create_engine(url)
        Session = sa_orm.sessionmaker()
        Session.configure(bind=engine)
        session = Session()

        # Define Session-based Metadata
        metadata = sa.MetaData(bind=session.bind)

        # table_name
        # metadata
        # engine
        """
          
    def get_url(self):
        
        # build the sqlalchemy URL
        self.url_string = URL.create(
        drivername=self.driver,
        host=self.REDSHIFT_HOST,# Amazon Redshift host
        port=self.REDSHIFT_PORT, # Amazon Redshift port
        database=self.REDSHIFT_DATABASE, # Amazon Redshift database
        username=self.REDSHIFT_USER, # Amazon Redshift username
        password=self.REDSHIFT_PASSWORD # Amazon Redshift password
        )

    def get_conection(self) -> None:

        self.logger.info("call get_conection() method of ControllerRedShift ")

        try:
                
                self.engine = sa.create_engine(self.url_string)
                Session = sa_orm.sessionmaker()
                Session.configure(bind=self.engine)
                self.session = Session()
                # Define Session-based Metadata
                self.metadata = sa.MetaData(bind=self.session.bind)
                
        except ClientError as err:
            self.logger.error('error  get_conection RedShift')
            raise TypeError('error  get_conection RedShift')

        self.logger.info(" Done get_conection() method of RedShift")

    def close_conection(self):          
        self.session.close() 
        self.engine.dispose()

    def execute_query(self,query=None):        

        return read_sql_query(text(query),self.engine)    

    def insert_dim_user(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_user (user_uid,postcode)
                        VALUES (%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['user_uid'],
                        dict_values['postcode']
                                )
            cur.execute(redshift_insert_query, record_to_insert)
            
    def insert_dim_account(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_account (account_uid,
                                                    bank_name,
                                                    account_type,
                                                    bank_type)
                        VALUES (%s,%s,%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['account_uid'],
                        dict_values['bank_name'],
                        dict_values['account_type'],
                        dict_values['bank_type']
                                )
            cur.execute(redshift_insert_query, record_to_insert)      
            
    def insert_dim_dates(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_dates (
                                                transaction_date,
                                                day_of_month,
                                                month_of_year,
                                                month_name,
                                                year,
                                                week_of_month,
                                                quarter_of_year
                                                )
                        VALUES (%s,%s,%s,%s,%s,%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['transaction_date'],
                        dict_values['day_of_month'],
                        dict_values['month_of_year'],
                        dict_values['month_name'],
                        dict_values['year'],
                        dict_values['week_of_month'],
                        dict_values['quarter_of_year']
                                )
            cur.execute(redshift_insert_query, record_to_insert)
            
    def date_descomposition(self,transaction_date):
        
        transaction_date = datetime.strptime(transaction_date, self.date_format).date()
        year = transaction_date.year
        month = transaction_date.month
        month_name = transaction_date.strftime('%b')
        day = transaction_date.day
        week_of_month = self.week_of_month(transaction_date)
        quarter_of_year = self.quarter_of_year(transaction_date)
        
        result = {
        'transaction_date':transaction_date,
        'day_of_month':day,
        'month_of_year':month,
        'month_name':month_name,
        'year':year,
        'week_of_month':week_of_month,
        'quarter_of_year':quarter_of_year
        }

        return result 
        
        # transaction_date.dt.weekday
        #transaction_date.dt.weekday_name

    def test(self):

        sql = "SELECT * FROM staging;"
        print(read_sql_query(text(sql),self.engine))
        
    def select_is_essential(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.nickname_recommended,
                    dim_transactions.is_essential_collective,
                    dim_transactions.is_essential_custom,
                    dim_transactions.is_essential_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.transaction_uid=dim_transactions.transaction_uid
                    AND
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        return read_sql_query(text(sql),self.engine)
    
    def select_is_membership(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.nickname_recommended,
                    dim_transactions.is_membership_collective,
                    dim_transactions.is_membership_custom,
                    dim_transactions.is_membership_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.transaction_uid=dim_transactions.transaction_uid
                    AND
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        return read_sql_query(text(sql),self.engine)
    
    def exists_user(self,user_uid):

        sql = f"""
                SELECT EXISTS (select 1 FROM dim_user
                WHERE dim_user.user_uid = '{user_uid}');
                """
        return read_sql_query(text(sql),self.engine)
      
    def select_nickname(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.clean_description,
                    dim_transactions.nickname_collective,
                    dim_transactions.nickname_custom,
                    dim_transactions.nickname_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.transaction_uid=dim_transactions.transaction_uid
                    AND
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        
        return read_sql_query(text(sql),self.engine)
    
    def select_subcategory(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.nickname_recommended,
                    dim_transactions.subcategory_uid_collective,
                    dim_transactions.subcategory_uid_custom,
                    dim_transactions.subcategory_uid_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.transaction_uid=dim_transactions.transaction_uid
                    AND
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        return read_sql_query(text(sql),self.engine)
    
    def exists(self, table_name,column_name ,value ):

        sql = "SELECT "+ column_name +" FROM " + table_name + " WHERE " + column_name + " = '"+ value +"';"
        
        if read_sql_query(text(sql),self.engine).shape[0] != 0:
            return True
        else : return False

    def week_of_month(self,dt):
        """ Returns the week of the month for the specified date.
        """

        first_day = dt.replace(day=1)

        dom = dt.day
        adjusted_dom = dom + first_day.weekday()

        return int(ceil(adjusted_dom/7.0))

    def quarter_of_year(self,dt):

        return ceil(dt.month/3.)

    def mapping_dict_(self,dict_values):
        return {
            'approved_asset' : dict_values['approved_asset'],
            'approved_business' : dict_values['approved_business'],
            'approved_category' : dict_values['approved_category'],
            'approved_frequency' : dict_values['approved_frequency'],
            'approved_income_source' : dict_values['approved_income_source'],
            'approved_is_essential' : dict_values['approved_is_essential'],
            'approved_is_loan_received' : dict_values['approved_is_loan_received'],
            'approved_is_loan_requested' : dict_values['approved_is_loan_requested'],
            'approved_is_membership' : dict_values['approved_is_membership'],
            'approved_nickname' : dict_values['approved_nickname'],
            'approved_person' : dict_values['approved_person'],
            'approved_special_moment' : dict_values['approved_special_moment'],
            'approved_subcategory' : dict_values['approved_subcategory'],
            'asset_uid_collective' : dict_values['asset_uid_def_collect'],
            'asset_uid_custom' : dict_values['asset_uid_def_usr'],
            'asset_uid_recommended' : dict_values['asset_uid_def_algo'],
            'business_uid_collective' : dict_values['business_uid_def_collect'],
            'business_uid_custom' : dict_values['business_uid_def_usr'],
            'business_uid_recommended' : dict_values['business_uid_def_algo'],
            'category_custom_type' : dict_values['category_custom_type'],
            'category_uid_collective' : dict_values['category_uid_def_collect'],
            'category_uid_custom' : dict_values['category_uid_def_usr'],
            'category_uid_recommended' : dict_values['category_uid_def_algo'],
            'classified_date' : dict_values['classified_date'],
            'clean_description' : dict_values['clean_description'],
            'country_code' : dict_values['country_code'],
            'created_homologation_date' : dict_values['created_homologation_date'],
            'flow_type' : dict_values['flow_type'],
            'frequency_uid_collective' : dict_values['frequency_uid_def_collect'],
            'frequency_uid_custom' : dict_values['frequency_uid_def_usr'],
            'frequency_uid_recommended' : dict_values['frequency_uid_def_algo'],
            'income_source_uid_collective' : dict_values['income_source_uid_def_collect'],
            'income_source_uid_custom' : dict_values['income_source_uid_def_usr'],
            'income_source_uid_recommended' : dict_values['income_source_uid_def_algo'],
            'is_essential_collective' : dict_values['is_essential_def_collect'],
            'is_essential_custom' : dict_values['is_essential_def_usr'],
            'is_essential_recommended' : dict_values['is_essential_def_algo'],
            'is_loan_received_collective' : dict_values['is_loan_received_def_collect'],
            'is_loan_received_custom' : dict_values['is_loan_received_def_usr'],
            'is_loan_received_recommended' : dict_values['is_loan_received_def_algo'],
            'is_loan_requested_collective' : dict_values['is_loan_requested_def_collect'],
            'is_loan_requested_custom' : dict_values['is_loan_requested_def_usr'],
            'is_loan_requested_recommended' : dict_values['is_loan_requested_def_algo'],
            'is_membership_collective' : dict_values['is_membership_def_collect'],
            'is_membership_custom' : dict_values['is_membership_def_usr'],
            'is_membership_recommended' : dict_values['is_membership_def_algo'],
            'language_code' : dict_values['language_code'],
            'message_type' : dict_values['message_type'],
            'model_classification_version' : dict_values['model_classification_version'],
            'model_collective_version' : dict_values['model_collective_version'],
            'model_nickname_version' : dict_values['model_nickname_version'],
            'model_nlp_version' : dict_values['model_nlp_version'],
            'nickname_collective' : dict_values['nickname_collect'],
            'nickname_custom' : dict_values['nickname_usr'],
            'nickname_recommended' : dict_values['nickname_algo'],
            'original_created_date' : dict_values['original_created_date'],
            'person_uid_collective' : dict_values['person_uid_def_collect'],
            'person_uid_custom' : dict_values['person_uid_def_usr'],
            'person_uid_recommended' : dict_values['person_uid_def_algo'],
            'source_description' : dict_values['source_description'],
            'special_moment_uid_collective' : dict_values['special_moment_uid_def_collect'],
            'special_moment_uid_custom' : dict_values['special_moment_uid_def_usr'],
            'special_moment_uid_recommended' : dict_values['special_moment_uid_def_algo'],
            'subcategory_custom_type' : dict_values['subcategory_custom_type'],
            'subcategory_uid_collective' : dict_values['subcategory_uid_def_collect'],
            'subcategory_uid_custom' : dict_values['subcategory_uid_def_usr'],
            'subcategory_uid_recommended' : dict_values['subcategory_uid_def_algo'],
            'transaction_uid' : dict_values['transaction_uid'],
            'updated_date' : dict_values['updated_at'],
            'validated' : dict_values['validated'],
            'validated_date' : dict_values['creted_at']                    
        }

    def insert_dim_transaction(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_transactions (
                        approved_asset,
                        approved_business,
                        approved_category,
                        approved_frequency,
                        approved_income_source,
                        approved_is_essential,
                        approved_is_loan_received,
                        approved_is_loan_requested,
                        approved_is_membership,
                        approved_nickname,
                        approved_person,
                        approved_special_moment,
                        approved_subcategory,
                        asset_uid_collective,
                        asset_uid_custom,
                        asset_uid_recommended,
                        business_uid_collective,
                        business_uid_custom,
                        business_uid_recommended,
                        category_custom_type,
                        category_uid_collective,
                        category_uid_custom,
                        category_uid_recommended,
                        classified_date,
                        clean_description,
                        country_code,
                        created_homologation_date,
                        flow_type,
                        frequency_uid_collective,
                        frequency_uid_custom,
                        frequency_uid_recommended,
                        income_source_uid_collective,
                        income_source_uid_custom,
                        income_source_uid_recommended,
                        is_autocategorized,
                        is_essential_collective,
                        is_essential_custom,
                        is_essential_recommended,
                        is_loan_received_collective,
                        is_loan_received_custom,
                        is_loan_received_recommended,
                        is_loan_requested_collective,
                        is_loan_requested_custom,
                        is_loan_requested_recommended,
                        is_membership_collective,
                        is_membership_custom,
                        is_membership_recommended,
                        language_code,
                        message_type,
                        model_classification_version,
                        model_collective_version,
                        model_nickname_version,
                        model_nlp_version,
                        nickname_collective,
                        nickname_custom,
                        nickname_recommended,
                        original_created_date,
                        person_uid_collective,
                        person_uid_custom,
                        person_uid_recommended,
                        source_description,
                        special_moment_uid_collective,
                        special_moment_uid_custom,
                        special_moment_uid_recommended,
                        subcategory_custom_type,
                        subcategory_uid_collective,
                        subcategory_uid_custom,
                        subcategory_uid_recommended,
                        transaction_uid,
                        updated_date,
                        validated,
                        validated_date
                                )
                        VALUES (
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s
                                )
                                    '''
            record_to_insert = (
                                 dict_values['approved_asset'],
                                dict_values['approved_business'],
                                dict_values['approved_category'],
                                dict_values['approved_frequency'],
                                dict_values['approved_income_source'],
                                dict_values['approved_is_essential'],
                                dict_values['approved_is_loan_received'],
                                dict_values['approved_is_loan_requested'],
                                dict_values['approved_is_membership'],
                                dict_values['approved_nickname'],
                                dict_values['approved_person'],
                                dict_values['approved_special_moment'],
                                dict_values['approved_subcategory'],
                                dict_values['asset_uid_def_collect'],
                                dict_values['asset_uid_def_usr'],
                                dict_values['asset_uid_def_algo'],
                                dict_values['business_uid_def_collect'],
                                dict_values['business_uid_def_usr'],
                                dict_values['business_uid_def_algo'],
                                dict_values['category_custom_type'],
                                dict_values['cat_uid_def_collect'],
                                dict_values['cat_uid_def_usr'],
                                dict_values['cat_uid_def_algo'],
                                dict_values['classified_at'],
                                dict_values['clean_description'],
                                dict_values['country_code'],
                                dict_values['created_at_homologation'],
                                dict_values['flow_type'],
                                dict_values['frequency_uid_def_collect'],
                                dict_values['frequency_uid_def_usr'],
                                dict_values['frequency_uid_def_algo'],
                                dict_values['income_source_uid_def_collect'],
                                dict_values['income_source_uid_def_usr'],
                                dict_values['income_source_uid_def_algo'],
                                dict_values['is_autocategorized'],
                                dict_values['is_essential_def_collect'],
                                dict_values['is_essential_def_usr'],
                                dict_values['is_essential_def_algo'],
                                dict_values['is_loan_received_def_collect'],
                                dict_values['is_loan_received_def_usr'],
                                dict_values['is_loan_received_def_algo'],
                                dict_values['is_loan_requested_def_collect'],
                                dict_values['is_loan_requested_def_usr'],
                                dict_values['is_loan_requested_def_algo'],
                                dict_values['is_membership_def_collect'],
                                dict_values['is_membership_def_usr'],
                                dict_values['is_membership_def_algo'],
                                dict_values['language_code'],
                                dict_values['message_type'],
                                dict_values['model_classification_version'],
                                dict_values['model_collective_version'],
                                dict_values['model_nickname_version'],
                                dict_values['model_nlp_version'],
                                dict_values['nickname_collect'],
                                dict_values['nickname_usr'],
                                dict_values['nickname_algo'],
                                dict_values['original_created_date'],
                                dict_values['person_uid_def_collect'],
                                dict_values['person_uid_def_usr'],
                                dict_values['person_uid_def_algo'],
                                dict_values['source_description'],
                                dict_values['special_moment_uid_def_collect'],
                                dict_values['special_moment_uid_def_usr'],
                                dict_values['special_moment_uid_def_algo'],
                                dict_values['subcategory_custom_type'],
                                dict_values['subcat_uid_def_collect'],
                                dict_values['subcat_uid_def_usr'],
                                dict_values['subcat_uid_def_algo'],
                                dict_values['transaction_uid'],
                                dict_values['updated_at'],
                                dict_values['validated'],
                                dict_values['created_at']                               
                                                       )
            cur.execute(redshift_insert_query, record_to_insert)

    def insert_dim_visualization(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_visualization (
                                asset_uid,
                                business_uid,
                                category_uid,
                                classified_date,
                                clean_description,
                                country_code,
                                created_homologation_date,
                                description,
                                flow_type,
                                frequency_uid,
                                is_autocategorized,
                                income_source_uid,
                                is_essential,
                                is_loan_received,
                                is_loan_requested,
                                is_membership,
                                language_code,
                                message_type,
                                original_created_date,
                                person_uid,
                                source_description,
                                special_moment_uid,
                                subcategory_uid,
                                transaction_uid,
                                updated_date,
                                validated,
                                validated_date
                                )
                        VALUES (
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s
                                )
                                    '''
            record_to_insert = (
                        dict_values['asset_uid'],
                        dict_values['business_uid'],
                        dict_values['category_uid'],
                        dict_values['classified_at'],
                        dict_values['clean_description'],
                        dict_values['country_code'],
                        dict_values['created_at_homologation'],
                        dict_values['description'],
                        dict_values['flow_type'],
                        dict_values['frequency_uid'],
                        dict_values['is_autocategorized'],
                        dict_values['income_source_uid'],
                        dict_values['is_essential'],
                        dict_values['is_loan_received'],
                        dict_values['is_loan_requested'],
                        dict_values['is_membership'],
                        dict_values['language_code'],
                        dict_values['message_type'],
                        dict_values['original_created_date'],
                        dict_values['person_uid'],
                        dict_values['source_description'],
                        dict_values['special_moment_uid'],
                        dict_values['subcategory_uid'],
                        dict_values['transaction_uid'],
                        dict_values['updated_at'],
                        dict_values['validated'],
                        dict_values['created_at']

                            )
            cur.execute(redshift_insert_query, record_to_insert)

    def insert_fact_transactions(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO fact_transactions (
                                                transaction_date,
                                                transaction_uid,
                                                user_uid,
                                                account_uid,
                                                amount,
                                                currency
                                                )
                        VALUES (%s,%s,%s,%s,%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['transaction_date'],
                        dict_values['transaction_uid'],
                        dict_values['user_uid'],
                        dict_values['account_uid'],
                        dict_values['amount'],
                        dict_values['currency']
                                )
            cur.execute(redshift_insert_query, record_to_insert)

    def insert_fact_visualization(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO fact_visualization (
                                                transaction_date,
                                                transaction_uid,
                                                user_uid,
                                                account_uid,
                                                amount,
                                                currency
                                                )
                        VALUES (%s,%s,%s,%s,%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['transaction_date'],
                        dict_values['transaction_uid'],
                        dict_values['user_uid'],
                        dict_values['account_uid'],
                        dict_values['amount'],
                        dict_values['currency']
                                )
            cur.execute(redshift_insert_query, record_to_insert)
            
    def insert_(self):
        """
        INSERT INTO target
        SELECT * FROM temp
        WHERE temp.primary_key NOT IN (SELECT DISTINCT primary_key FROM target)
        """

    def creted_table(self,table_name = 'staging'):
        
        RedshiftDBTable = sa.Table(
                                table_name,
                                self.metadata,
                                sa.Column('session_id', sa.VARCHAR(80)),
                                sa.Column('click_region', sa.VARCHAR(100)),
                                sa.Column('product_id', sa.VARCHAR(40)),
                                redshift_diststyle='KEY',
                                redshift_distkey='session_id',
                                redshift_sortkey='product_id'
                                    )

        # Drop the table if it already exists
        if sa.inspect(self.engine).has_table(table_name):
            RedshiftDBTable.drop(bind=self.engine)

        # Create the table (execute the "CREATE TABLE" SQL statement for "")
        RedshiftDBTable.create(bind=self.engine)
        self.session.commit()
        
    def get_date_range(self,num_days = 30):
        # today 
        to_date = datetime.utcnow()
        from_date = to_date  - timedelta(days=num_days)

        return from_date, to_date

    def get_collective_is_essential(self,num_perceltin_threshold=9):

        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_is_essential = self.select_is_essential(start = from_date, end = to_date).dropna(subset=['nickname_recommended'])

        df_is_essential = df_is_essential[df_is_essential.nickname_recommended != "No Suggestion"]
        cols = [
                'is_essential_collective',
                'is_essential_custom',
                'is_essential_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_is_essential[['transaction_uid','nickname_recommended', col]]
            df_tmp.rename(columns={col: "is_essential"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['nickname_recommended','is_essential']).count()['transaction_uid'] \
                                   .groupby(['nickname_recommended','is_essential']).agg('max'))
        
        threshold = percentile(result['transaction_uid'].values, arange(0, 100, 10))[num_perceltin_threshold] 
        result = result[result.transaction_uid >  threshold]
        
        out = (result.reset_index(level=1).groupby(level=[0])[['is_essential']]
       .apply(lambda x: x.to_dict('records'))#
       .reset_index()
       .rename(columns={0:'config'})
        .to_json(orient='values')
            )

        out = loads_(out)

        dict_ = dict()
        for key,config in out :
            dict_[key] = config[0]

        return  dict_
    # result.reset_index().drop_duplicates(subset='nickname_recommended', keep='first').set_index('nickname_recommended')['is_essential'].to_dict()

    def get_collective_is_membership(self,num_perceltin_threshold =9):

        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_is_membership = self.select_is_membership(start = from_date, end = to_date).dropna(subset=['nickname_recommended'])

        df_is_membership = df_is_membership[df_is_membership.nickname_recommended != "No Suggestion"]
        cols = [
                'is_membership_collective',
                'is_membership_custom',
                'is_membership_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_is_membership[['transaction_uid','nickname_recommended', col]]
            df_tmp.rename(columns={col: "is_membership"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['nickname_recommended','is_membership']).count()['transaction_uid'] \
                                   .groupby(['nickname_recommended','is_membership']).agg('max'))
        
        threshold = percentile(result['transaction_uid'].values, arange(0, 100, 10))[num_perceltin_threshold] 
        result = result[result.transaction_uid >  threshold]

        out = (result.reset_index(level=1).groupby(level=[0])[['is_membership']]
       .apply(lambda x: x.to_dict('records'))#
       .reset_index()
       .rename(columns={0:'config'})
        .to_json(orient='values')
            )

        out = loads_(out)

        dict_ = dict()
        for key,config in out :
            dict_[key] = config[0]

        return  dict_
    
        # return result.reset_index().drop_duplicates(subset='nickname_recommended', keep='first').set_index('nickname_recommended')['is_membership'].to_dict()

    def get_collective_nickname(self, num_perceltin_threshold = 9):

        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_nickname = self.select_nickname(start = from_date, end = to_date).dropna(subset=['clean_description'])

        df_nickname = df_nickname[df_nickname.nickname_recommended != "No Suggestion"]
        
        cols = [
                'nickname_collective',
                'nickname_custom',
                'nickname_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_nickname[['transaction_uid','clean_description', col]]
            df_tmp.rename(columns={col: "nickname"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['clean_description','nickname']).count()['transaction_uid'] \
                                   .groupby(['clean_description','nickname']).agg('max'))
        threshold = percentile(result['transaction_uid'].values, arange(0, 100, 10))[num_perceltin_threshold] 
        result = result[result.transaction_uid >  threshold]
        
        out = (result.reset_index(level=1).groupby(level=[0])[['nickname']]
       .apply(lambda x: x.to_dict('records'))#
       .reset_index()
       .rename(columns={0:'config'})
        .to_json(orient='values')
             )

        out = loads_(out)

        dict_ = dict()
        for key,config in out :
            dict_[key] = config[0]
        return dict_
        #return result.reset_index().drop_duplicates(subset='clean_description', keep='first').set_index('clean_description')['nickname'].to_dict()

    def get_collective_subcategory(self,num_perceltin_threshold=9):
        
        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_subcategory = self.select_subcategory(start = from_date, end = to_date).dropna(subset=['nickname_recommended'])

        df_subcategory = df_subcategory[df_subcategory.nickname_recommended != "No Suggestion"]
        
        cols = [
            'subcategory_uid_collective',
            'subcategory_uid_custom',
            'subcategory_uid_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_subcategory[['transaction_uid','nickname_recommended', col]]
            df_tmp.rename(columns={col: "subcategory_uid"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['nickname_recommended','subcategory_uid']).count()['transaction_uid'] \
                                   .groupby(['nickname_recommended','subcategory_uid']).agg('max'))
        
        threshold = percentile(result['transaction_uid'].values, arange(0, 100, 10))[num_perceltin_threshold] 
        result = result[result.transaction_uid >  threshold]
        
        out = (result.reset_index(level=1).groupby(level=[0])[['subcategory_uid']]
       .apply(lambda x: x.to_dict('records'))#
       .reset_index()
       .rename(columns={0:'config'})
        .to_json(orient='values')
            )

        out = loads_(out)

        dict_ = dict()
        for key,config in out :
            dict_[key] = config[0]

        return  dict_
        # return result.reset_index().drop_duplicates(subset='nickname_recommended', keep='first').set_index('nickname_recommended')['subcategory_uid'].to_dict()



import time
from math import ceil

from numpy import unique
from os import getenv
from datetime import datetime
from itertools import islice

from pykafka import KafkaClient
from pykafka.common import OffsetType
from klp_commons.log.log import Log
from klp_commons.log.structuredMessage import StructuredMessage
from klp_commons.controllers.controller_mongodb import ControllerMongoDB

from traceback import  format_exc
from sys import exc_info, exit

null=None; true=True; false=False
fmt = StructuredMessage

from kafka import KafkaProducer
from json import dumps as dumps_ 
from json import  loads as loads_

from os import getenv
from pykafka import KafkaClient
from pykafka.common import OffsetType
import math
from itertools import islice
import json
from datetime import datetime

null=None; true=True; false=False

from kafka import KafkaProducer
from json import dumps as dumps_


class Producer:
    def __init__(
                self, 
                topic = None,
                kafka_brokers = None,
                acks = None,
                api_version = None,
                client_id = None
                 ):
        self.kafka_api_version = api_version
        self.topic = topic
        self.acks = acks
        self.client_id = client_id
        self.kafka_brokers = kafka_brokers
        

    def init_producer(self) -> None:

        self.producer = KafkaProducer(
            acks = self.acks,
            api_version=self.kafka_api_version,
            bootstrap_servers = self.kafka_brokers,
            value_serializer = lambda x: dumps_(x).encode('utf-8'),
        )

    def send_message(self, data: dict) -> None:
        self.producer.send(self.topic, value=data)
        # block until all async messages are sent
        self.producer.flush()

class Topics():

    def __init__(self,topics,
                level: str = 'info', # logging levels
                 event_level: str = 'error', # event logging levels  for sentry
                 filename_log: str = 'topis.log', # file
                 is_local_env: bool = False # enable Sentry 
                 ): 
        self.topics = topics
        self.kafka_brokers = None
        self.earliest_available_offsets = None
        self.latest_offsets = None
        self.list_type_messages = []
        self.list_content_messages = []
        self.others_types_messages_list = ['machine-learning-asset-created',
                                'machine-learning-asset-deleted',
                                'machine-learning-asset-updated',
                                'machine-learning-business-created',
                                'machine-learning-business-deleted',
                                'machine-learning-business-updated',
                                'machine-learning-person-created',
                                'machine-learning-person-deleted',
                                'machine-learning-person-updated',
                                'machine-learning-special-moment-created',
                                'machine-learning-special-moment-deleted',
                                'machine-learning-special-moment-updated',
                                'machine-learning-expense-created',
                                'machine-learning-expense-updated',
                                'machine-learning-expense-deleted',
                                'machine-learning-income-created',
                                'machine-learning-income-updated',
                                'custom-category-created',
                                'custom-category-updated',
                                ]
        self.MongoDB = ControllerMongoDB()
        
        if filename_log is None:
            filename_log = 'Topics.log'

        self.Logger = Log(
            path_file = filename_log,
            level = level,
            event_level = event_level,
            is_local_env = is_local_env)

        self.Logger.startlogging()

        # module_name.TopicKafka_name
        self.logger = self.Logger.get_log('Topics')
        self.logger.info("Constructor Topics()")

        self.set_client_kafka()

    def set_client_kafka(self):
        
        port_current = '9094'
        port_new = '9092'
        default_kafka_host = 'localhost:9092'
        enable_port_replace = True
        self.acks = 'all'
        self.api_version = (2,6,1)
        self.group_id  = 'categorization-update-kw-group'
        self.client_id = None

        if getenv('KAFKA_BROKERS'):
            if len(getenv('KAFKA_BROKERS')) != 0:
                self.kafka_brokers = getenv('KAFKA_BROKERS')

        elif getenv('KAFKA_HOST') and getenv('KAFKA_PORT'):
            self.kafka_brokers = getenv('KAFKA_HOST') + ':' + getenv('KAFKA_PORT')
        else :
            self.kafka_brokers = default_kafka_host

        if enable_port_replace:
            self.kafka_brokers = self.kafka_brokers.replace(port_current,port_new)

        self.client = KafkaClient(hosts=self.kafka_brokers)
       
    
    def extract_offset_index(self):
        """
        Obtener el offset más antiguo disponible (earliest) y el 
        offset más reciente (latest) de la lista de tópicos
        almacenada en la variable topics
        """
        self.logger.info("Call get_offset_index()")

        self.latest_offsets = {k: 0 for k in self.topics}
        self.earliest_available_offsets = {k: 0 for k in self.topics}

        for topic_ in self.topics:
            self.logger.info(fmt('Get index for: ', json={"topic":topic_}))

            topic = self.client.topics[topic_]
            self.earliest_available_offsets[topic_] = topic.earliest_available_offsets()[0][0][0]
            self.latest_offsets[topic_] = topic.latest_available_offsets()[0][0][0]

        self.logger.info("Done call get_offset_index()")
        
    def get_offsets(self):
        return self.earliest_available_offsets, self.latest_offsets  
    
    def extract_messages(self,topic_,last_num_messages,flag_content = False, flag_send= False):
    
    
        topic = self.client.topics[topic_]


        consumer = topic.get_simple_consumer(auto_offset_reset=OffsetType.LATEST,reset_offset_on_start=True)


        # how many messages should we get from the end of each partition?
        MAX_PARTITION_REWIND = int(ceil(last_num_messages / len(consumer._partitions)))

        # find the beginning of the range we care about for each partition
        offsets = [(p, op.last_offset_consumed - MAX_PARTITION_REWIND) for p, op in consumer._partitions.items()]

        # if we want to rewind before the beginning of the partition, limit to beginning
        offsets = [(p, (o if o > -1 else -2)) for p, o in offsets]

        # reset the consumer's offsets
        consumer.reset_offsets(offsets)

        message_count = 0
        message_categorized_count = 0
        for message in islice(consumer, last_num_messages):
            self.logger.info(fmt('Processing ... current offset ', json={"offset":message.offset}))

            # same time of all variables crete_at/update value 

            self.dt =  datetime.utcnow().isoformat()
            # self.dt = datetime.now(timezone.utc).strftime(self.format_inter)
            self.offset = message.offset



            self.logger.info("message processing ... ")
            
            # encrypted = Crypto.encrypt(data)

            current_offset = self.offset
            message = loads_(message.value)
            
            self.logger.info(fmt("message type and offset incoming", json={'type':message['type'],"offset":self.offset}))
            print("type:", message['type'], "  offset: ", self.offset)
            
            
            # message['offset_'] = self.offset
            # message['datetime_reporting'] = self.dt
            message_count += 1

            self.list_type_messages.append(message["type"])

            if flag_content:
                self.list_content_messages.append(message)
            if flag_send:
                self.send_message(message)

            if message_count == last_num_messages :
                self.logger.info("message_count == last_num_messages -> True")
                self.logger.info("Break Loop")
                consumer.stop()
                time.sleep(0.1)
                break

    def get_messages(self):
        return self.list_type_messages, self.list_content_messages,
                
    def value_count_type_message(self):

        values, counts = unique(self.list_type_messages, return_counts=True)

        return values, counts

    def resend(self, save_mongo = False, start_time = None, end_time = None):

        self.extract_offset_index()

        earliest_available_offsets, latest_offsets = self.get_offsets()
        print("earliest_available_offsets: ", earliest_available_offsets)
        print("latest_offsets: ", latest_offsets)

        start_time = datetime.strptime(start_time, '%Y-%m-%d').date()
        end_time = datetime.strptime(end_time, '%Y-%m-%d').date()

        for topic in self.topics :
            print("topic: ", topic)
            last_num_messages = latest_offsets[topic] - earliest_available_offsets[topic]

            self.extract_messages(topic,last_num_messages,flag_content = True,flag_send = save_mongo)

            set_of_jsons = {json.dumps(d, sort_keys=True) for d in self.list_content_messages}
            list_content_messages = [json.loads(t) for t in set_of_jsons]

            print("Remover duplicados... \nNúmero de transacciones: ",len(list_content_messages))      
            
            print("list_content_messages: ", list_content_messages)
            """
            list_final = list()
            for message in list_content_messages:
                if 'machine-learning' in message['type']:
                    datetime_str = message['data'][0]['created_at'][0:10]
                    print(datetime_str)
                    datetime_object = datetime.strptime(datetime_str, '%Y-%m-%d').date()
                    if datetime_object >= start_time and datetime_object<=end_time:
                        list_final.append(message)
            
            print("list_final: ", list_final)
            """
            # ==========================
            self.ProdTrans = Producer(topic, self.kafka_brokers, self.acks, self.api_version , self.client_id)
            self.ProdTrans.init_producer()

            for item in list_content_messages:
                self.ProdTrans.send_message(item)
            
    def value_counts_types(self):
        
        dict_value_counts = dict()
        
        for topic in self.topics :
            print("topic: ", topic)
            last_num_messages = latest_offsets[topic] - earliest_available_offsets[topic]

            self.extract_messages(topic,last_num_messages)

            values, counts = unique(self.list_type_messages, return_counts=True)
            dict_value_counts[topic] = zip(values, counts)
        
        return dict_value_counts
            
    def send_message(self,message):

        try :
            if message['type'] == 'transaction-homologated':
                self.logger.info("Call save transaction-homologated ")

                message['_id'] = self.offset
                self.MongoDB.get_con_collect(collection_name='transaction-homologated')
                self.MongoDB.set_document(message)
                # fix insert many 
                self.MongoDB.insert_one_collection()
                self.MongoDB.close_con()
                self.logger.info("Don call save transaction-homologated ")
            elif message['type'] in self.others_types_messages_list :
                
                self.logger.info("Call save into " + message['type'] + " colecction mongodb ")

                if message['type'] == 'machine-learning-asset-created':
                    for msj in message['data']['assetsCreated']:
                        msj['_id'] = msj['uid']
                        del msj['uid']

                        self.MongoDB.get_con_collect(collection_name=message['type'].split("machine-learning-")[1])
                        self.MongoDB.set_document(msj)
                        # fix insert many 
                        self.MongoDB.insert_one_collection()
                        self.MongoDB.close_con()

                elif isinstance(message['data'] , list) :
                    for msj in message['data']:

                        msj['_id'] = msj['uid']
                        del msj['uid']

                        self.MongoDB.get_con_collect(collection_name=message['type'].split("machine-learning-")[1])
                        self.MongoDB.set_document(msj)
                        # fix insert many 
                        self.MongoDB.insert_one_collection()
                        self.MongoDB.close_con()


                else :

                    message['_id'] = message['data']['uid']
                    del message['data']['uid']
                    self.MongoDB.get_con_collect(collection_name=message['type'].split("machine-learning-")[1])

                    self.MongoDB.set_document(message)
                    self.MongoDB.insert_one_collection()
                    self.MongoDB.close_con()

        except ValueError as e:
            exit
            traceback_info = format_exc()
            type_ , value , _ = exc_info()
            
            json_ = {
                    'type': message['type'],
                    'error_type':str(type_.__name__),
                    'error_value':str(value),
                    'traceback_info': traceback_info
                    }
            
            self.logger.error(fmt('Exception for processing message. Save message traceback info, details error:', json = json_))
        
        
        
        self.logger.info("Done save into " + message['type'] + " colecction mongodb ")




if __name__ == "__main__":

    list_topics = ['transaction','machine-learning'] 
    level_default = 'info'
    event_level_default = 'error'
    fn = 'topics.log'
    save_db = True

    topics = Topics(list_topics,
                    level=level_default,
                    event_level=event_level_default,
                    filename_log=fn,
                    is_local_env=False
                   )

    topics.extract_offset_index()


    earliest_available_offsets, latest_offsets = topics.get_offsets()
    
    print(earliest_available_offsets, "\n _:_ \n" ,latest_offsets)


    if save_db:
        topics.save_messages_to_mongodb()


    print(topics.value_count_type_message())

    print("Finish script OK ")




import base64
from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad,unpad
import base64
from hashlib import md5
from os import getenv

class CryptologyJS:

    def __init__(self,pad : int = None):
        self.BLOCK_SIZE = 16
        self.password = getenv("JWT_SECRET").encode()        
    
    def pad(self,data):
        length = self.BLOCK_SIZE - (len(data) % self.BLOCK_SIZE)
        return data + (chr(length)*length).encode()

    def unpad(self,data):
        return data[:-(data[-1] if type(data[-1]) == int else ord(data[-1]))]
    
    def bytes_to_key(self,data, salt, output=48):
    # extended from https://gist.github.com/gsakkis/4546068
        assert len(salt) == 8, len(salt)
        data += salt
        key = md5(data).digest()
        final_key = key
        while len(final_key) < output:
            key = md5(key + data).digest()
            final_key += key
        return final_key[:output]

    def encrypt(self, message):
        salt = Random.new().read(8)
        key_iv = self.bytes_to_key(self.password, salt, 32+16)
        key = key_iv[:32]
        iv = key_iv[32:]
        aes = AES.new(key, AES.MODE_CBC, iv)
        return base64.b64encode(b"Salted__" + salt + aes.encrypt(self.pad(message)))

    def decrypt(self, message):
        message = base64.b64decode(message)
        assert message[0:8] == b"Salted__"
        salt = message[8:16]
        key_iv = self.bytes_to_key(self.password, salt, 32+16)
        key = key_iv[:32]
        iv = key_iv[32:]
        aes = AES.new(key, AES.MODE_CBC, iv)
        return self.unpad(aes.decrypt(message[16:]))




from numpy import ndarray
from numpy import floating
from numpy import integer
from json import JSONEncoder

def divide_chunks(l: list = None, n: int = 100) -> list:

    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]

def get_full_class_name(obj):
    """
    get full class name error 
    """
    module = obj.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return obj.__class__.__name__
    return module + '.' + obj.__class__.__name__

class CustomEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, integer):
            return int(obj)
        elif isinstance(obj, floating):
            return float(obj)
        elif isinstance(obj, ndarray):
            return obj.tolist()
        else:
            return super(CustomEncoder, self).default(obj)

